/*
 * ble_zive.h
 *
 *  Created on: Oct 18, 2018
 *      Author: miceuz
 */

#ifndef BLE_ZIVE_H_
#define BLE_ZIVE_H_

/**@brief   Macro for defining a ble_lbs instance.
 *
 * @param   _name   Name of the instance.
 * @hideinitializer
 */

#include <stdint.h>
#include <stdbool.h>
#include "ble_srv_common.h"
#include "nrf_sdh_ble.h"
#include "nrf_ble_gatt.h"
#include "ble_gatts.h"
#include "ble.h"

#ifdef __cplusplus
extern "C" {
#endif

#define BLE_ZIVE_MAX_BUFFERED_ECG_RECORDS 100


#define ZIVE_UUID_BASE        {0x8e, 0xad, 0x84, 0x84, 0x37, 0x69, 0x4c, 0xc9, \
                              0x89, 0x46, 0xc7, 0x8c, 0xf4, 0xa0, 0x3d, 0xb8}
#define ZIVE_UUID_SERVICE     0x4223
#define ZIVE_UUID_DATA_CHAR   0x4224
#define ZIVE_UUID_CONFIG_CHAR 0x4225


#define BLE_ZIVE_DEF(_name)                                                                          \
static ble_zive_t _name;                                                                             \
NRF_SDH_BLE_OBSERVER(_name ## _obs,                                                                 \
                     BLE_LBS_BLE_OBSERVER_PRIO,                                                     \
                     ble_zive_on_ble_evt, &_name)

/**@brief ZiveService event type. */
typedef enum {
    BLE_ZIVE_EVT_NOTIFICATION_ENABLED,   /**< Notification enabled event. */
    BLE_ZIVE_EVT_NOTIFICATION_DISABLED,   /**< Notification disabled event. */
    BLE_ZIVE_EVT_DISCONNECT   /**< Notification disabled event. */
} ble_zive_evt_type_t;

/**@brief Zive Rate Service event. */
typedef struct {
    ble_zive_evt_type_t evt_type;    /**< Type of event. */
} ble_zive_evt_t;

// Forward declaration of the ble_zive_t type.
typedef struct ble_zive_s ble_zive_t;

/**@brief Heart Rate Service event handler type. */
typedef void (*ble_zive_evt_handler_t) (ble_zive_t * p_zive, ble_zive_evt_t * p_evt);

/**@brief Zive Service init structure. This contains all options and data needed for
 *        initialization of the service. */
typedef struct {
    ble_zive_evt_handler_t        evt_handler;                                          /**< Event handler to be called for handling events in the Heart Rate Service. */
    security_req_t               data_cccd_wr_sec;                                      /**< Security requirement for writing the DATA characteristic CCCD. */
    security_req_t               ctrl_rd_sec;                                           /**< Security requirement for reading the CTRL characteristic value. */
} ble_zive_init_t;

/**@brief Zive Service structure. This contains various status information for the service. */
struct ble_zive_s
{
    ble_zive_evt_handler_t        evt_handler;                                          /**< Event handler to be called for handling events in the Heart Rate Service. */
    uint16_t                     service_handle;                                       /**< Handle of Zive Service (as provided by the BLE stack). */
    ble_gatts_char_handles_t     data_handles;                                          /**< Handles related to the Data characteristic. */
    ble_gatts_char_handles_t     ctrl_handles;                                          /**< Handles related to the Control characteristic. */
    uint16_t                     conn_handle;                                          /**< Handle of the current connection (as provided by the BLE stack, is BLE_CONN_HANDLE_INVALID if not in a connection). */
    bool                         is_sensor_contact_detected;                           /**< TRUE if sensor contact has been detected. */
//    uint16_t                     ecg_data[BLE_ZIVE_MAX_BUFFERED_ECG_RECORDS];               /**< Set of ECG data points since the last transmission. */
//    uint16_t                     rr_interval_count;                                    /**< Number of ECG data points since the last transmission. */
    uint8_t                      max_data_len;                                          /**< Current maximum ECG measurement length, adjusted according to the current ATT MTU. */
    uint8_t                     uuid_type;           /**< UUID type for the Zive Button Service. */
};

/**@brief Function for initializing the Zive Service.
 *
 * @param[out]  p_zive       Zive Service structure. This structure will have to be supplied by
 *                           the application. It will be initialized by this function, and will later
 *                           be used to identify this particular service instance.
 * @param[in]   p_zive_init  Information needed to initialize the service.
 *
 * @return      NRF_SUCCESS on successful initialization of service, otherwise an error code.
 */
uint32_t ble_zive_init(ble_zive_t * p_zive, ble_zive_init_t const * p_zive_init);


/**@brief Function for handling the GATT module's events.
 *
 * @details Handles all events from the GATT module of interest to the Zive Service.
 *
 * @param[in]   p_zive       Zive Service structure.
 * @param[in]   p_gatt_evt  Event received from the GATT module.
 */
void ble_zive_on_gatt_evt(ble_zive_t * p_zive, nrf_ble_gatt_evt_t const * p_gatt_evt);


/**@brief Function for handling the Application's BLE Stack events.
 *
 * @details Handles all events from the BLE stack of interest to the Zive Service.
 *
 * @param[in]   p_ble_evt   Event received from the BLE stack.
 * @param[in]   p_context   Zive Service structure.
 */
void ble_zive_on_ble_evt(ble_evt_t const * p_ble_evt, void * p_context);


/**@brief Function for sending data chunk if notification has been enabled.
 *
 * @details The application calls this function after having collected ECG data.
 *          If notification has been enabled, the ECG data is encoded and sent to
 *          the client.
 *
 * @param[in]   p_zive                   Zive Service structure.
 * @param[in]   p_data                   ECG data to send.
 * @param[in]   length                   Length of ECG data to send.
 *
 * @return      NRF_SUCCESS on success, otherwise an error code.
 */
uint32_t ble_zive_ecg_send(ble_zive_t * p_zive, int32_t * p_data, uint16_t length);

/**@brief Function for setting the state of the Sensor Contact Detected bit.
 *
 * @param[in]   p_zive                        Heart Rate Service structure.
 * @param[in]   is_sensor_contact_detected   TRUE if sensor contact is detected, FALSE otherwise.
 */
void ble_zive_sensor_contact_detected_update(ble_zive_t * p_zive, bool is_sensor_contact_detected);




#ifdef __cplusplus
}
#endif

#endif /* BLE_ZIVE_H_ */
